@@ -6,15 +6,13 @@
 # Public License as described in the bup LICENSE file.
 
 from __future__ import absolute_import, print_function
-from binascii import hexlify
 from copy import deepcopy
 from errno import EACCES, EINVAL, ENOTTY, ENOSYS, EOPNOTSUPP
 from io import BytesIO
 from time import gmtime, strftime
-import errno, os, sys, stat, time, pwd, grp, socket, struct
+import errno, os, sys, stat, time, socket, struct
 
-from bup import compat, vint, xstat
-from bup.compat import py_maj
+from bup import vint, xstat
 from bup.drecurse import recursive_dirlist
 from bup.helpers import add_error, mkdirp, log, is_superuser, format_filesize
 from bup.io import path_msg
@@ -189,13 +187,14 @@ _rec_tag_end = 0
 _rec_tag_path = 1
 _rec_tag_common_v1 = 2 # times, user, group, type, perms, etc. (legacy/broken)
 _rec_tag_symlink_target = 3
-_rec_tag_posix1e_acl = 4      # getfacl(1), setfacl(1), etc.
+_rec_tag_posix1e_acl_v1 = 4   # (broken \n delimited format, see v2 below)
 _rec_tag_nfsv4_acl = 5        # intended to supplant posix1e? (unimplemented)
 _rec_tag_linux_attr = 6       # lsattr(1) chattr(1)
 _rec_tag_linux_xattr = 7      # getfattr(1) setfattr(1)
 _rec_tag_hardlink_target = 8 # hard link target path
 _rec_tag_common_v2 = 9 # times, user, group, type, perms, etc. (current)
 _rec_tag_common_v3 = 10  # adds optional size to v2
+_rec_tag_posix1e_acl_v2 = 11     # getfacl(1), setfacl(1), etc.
 
 _warned_about_attr_einval = None
 
@@ -455,8 +454,13 @@ class Metadata:
         if _have_lchmod:
             try:
                 os.lchmod(path, stat.S_IMODE(self.mode))
-            except errno.ENOSYS:  # Function not implemented
-                pass
+            except OSError as e:
+                # - "Function not implemented"
+                # - "Operation not supported" might be generated by glibc
+                if e.errno in (errno.ENOSYS, errno.EOPNOTSUPP):
+                    pass
+                else:
+                    raise
         elif not stat.S_ISLNK(self.mode):
             os.chmod(path, stat.S_IMODE(self.mode))
 
@@ -479,6 +483,9 @@ class Metadata:
         try:
             if stat.S_ISLNK(st.st_mode):
                 self.symlink_target = os.readlink(path)
+                # might have read a different link than the
+                # one that was in place when we did stat()
+                self.size = len(self.symlink_target)
         except OSError as e:
             add_error('readlink: %s' % e)
 
@@ -539,10 +546,36 @@ class Metadata:
         else:
             return None
 
-    def _load_posix1e_acl_rec(self, port):
+    @staticmethod
+    def _correct_posix1e_v1_delimiters(acls, path):
+        assert acls
+        # The v0 format had newline delimiters which are incorrect for
+        # the ACL short text format we request, and which are rejected
+        # with EINVAL by acl_from_text() on some platforms.  For now,
+        # this function assumes (potentially incorrectly) that no
+        # field name (including the user and group names) contains a
+        # newline or comma.  If any field name does, then the results
+        # may be wrong.  (Debian, at least, disallows them.)
+        for i in range(len(acls)):
+            acl = acls[i]
+            if b',' in acl:
+                if path:
+                    msg = f'Unexpected comma in ACL entry; ignoring {acl!r}' \
+                        f' for {path_msg(path)}\n'
+                else:
+                    msg = f'Unexpected comma in ACL entry; ignoring {acl!r}\n'
+                add_error(msg)
+                return None
+            acls[i] = acl.replace(b'\n', b',')
+        return acls
+
+    def _load_posix1e_acl_rec(self, port, *, version):
+        assert version in (1, 2)
         acl_rep = vint.unpack('ssss', vint.read_bvec(port))
         if acl_rep[2] == b'':
             acl_rep = acl_rep[:2]
+        if version == 1:
+            acl_rep = self._correct_posix1e_v1_delimiters(acl_rep, self.path)
         self.posix1e_acl = acl_rep
 
     def _apply_posix1e_acl_rec(self, path, restore_numeric_ids=False):
@@ -567,7 +600,7 @@ class Metadata:
                 # (or group) doesn't exist
                 raise ApplyError("POSIX1e ACL: can't create %r for %r"
                                  % (acls, path_msg(path)))
-            elif e.errno == errno.EPERM or e.errno == errno.EOPNOTSUPP:
+            elif e.errno in (errno.EPERM, errno.EOPNOTSUPP):
                 raise ApplyError('POSIX1e ACL applyto: %s' % e)
             else:
                 raise
@@ -689,8 +722,7 @@ class Metadata:
                 try:
                     xattr.set(path, k, v, nofollow=True)
                 except IOError as e:
-                    if e.errno == errno.EPERM \
-                            or e.errno == errno.EOPNOTSUPP:
+                    if e.errno in (errno.EPERM, errno.EOPNOTSUPP):
                         raise ApplyError('xattr.set %r: %s' % (path_msg(path), e))
                     else:
                         raise
@@ -705,6 +737,9 @@ class Metadata:
                     raise
 
     def __init__(self):
+        __slots__ = ('mode', 'uid', 'atime', 'mtime', 'ctime',
+                     'path', 'size', 'symlink_target', 'hardlink_target',
+                     'linux_attr', 'linux_xattr', 'posix1e_acl')
         self.mode = self.uid = self.gid = self.user = self.group = None
         self.atime = self.mtime = self.ctime = None
         # optional members
@@ -782,25 +817,25 @@ class Metadata:
         return ''.join(result)
 
     def write(self, port, include_path=True):
+        port.write(self.encode(include_path=include_path))
+
+    def encode(self, include_path=True):
+        ret = []
         records = include_path and [(_rec_tag_path, self._encode_path())] or []
         records.extend([(_rec_tag_common_v3, self._encode_common()),
                         (_rec_tag_symlink_target,
                          self._encode_symlink_target()),
                         (_rec_tag_hardlink_target,
                          self._encode_hardlink_target()),
-                        (_rec_tag_posix1e_acl, self._encode_posix1e_acl()),
+                        (_rec_tag_posix1e_acl_v2, self._encode_posix1e_acl()),
                         (_rec_tag_linux_attr, self._encode_linux_attr()),
                         (_rec_tag_linux_xattr, self._encode_linux_xattr())])
         for tag, data in records:
             if data:
-                vint.write_vuint(port, tag)
-                vint.write_bvec(port, data)
-        vint.write_vuint(port, _rec_tag_end)
-
-    def encode(self, include_path=True):
-        port = BytesIO()
-        self.write(port, include_path)
-        return port.getvalue()
+                ret.extend((vint.encode_vuint(tag),
+                            vint.encode_bvec(data)))
+        ret.append(vint.encode_vuint(_rec_tag_end))
+        return b''.join(ret)
 
     def copy(self):
         return deepcopy(self)
@@ -828,8 +863,10 @@ class Metadata:
                     result._load_symlink_target_rec(port)
                 elif tag == _rec_tag_hardlink_target:
                     result._load_hardlink_target_rec(port)
-                elif tag == _rec_tag_posix1e_acl:
-                    result._load_posix1e_acl_rec(port)
+                elif tag == _rec_tag_posix1e_acl_v2:
+                    result._load_posix1e_acl_rec(port, version=2)
+                elif tag == _rec_tag_posix1e_acl_v1:
+                    result._load_posix1e_acl_rec(port, version=1)
                 elif tag == _rec_tag_linux_attr:
                     result._load_linux_attr_rec(port)
                 elif tag == _rec_tag_linux_xattr:
@@ -884,13 +921,16 @@ class Metadata:
 
 def from_path(path, statinfo=None, archive_path=None,
               save_symlinks=True, hardlink_target=None,
-              normalized=False):
+              normalized=False, after_stat=None):
+    # This function is also a test hook; see test-save-errors
     """Return the metadata associated with the path.  When normalized is
     true, return the metadata appropriate for a typical save, which
     may or may not be all of it."""
     result = Metadata()
     result.path = archive_path
     st = statinfo or xstat.lstat(path)
+    if after_stat:
+        after_stat(path)
     result._add_common(path, st)
     if save_symlinks:
         result._add_symlink_target(path, st)
